/*
 * Copyright (C) 2017 FHNW
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ch.fhnw.lernstickwelcome.model.backup;

import ch.fhnw.lernstickwelcome.controller.exception.ProcessingException;
import ch.fhnw.lernstickwelcome.model.Processable;
import ch.fhnw.lernstickwelcome.model.WelcomeConstants;
import ch.fhnw.lernstickwelcome.model.WelcomeModelFactory;
import ch.fhnw.lernstickwelcome.util.WelcomeUtil;
import ch.fhnw.util.Partition;
import ch.fhnw.util.ProcessExecutor;
import ch.fhnw.util.StorageDevice;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.concurrent.Task;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.freedesktop.dbus.exceptions.DBusException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * This class handles configurations for the backup process.
 *
 * @author sschw
 */
public class BackupTask implements Processable<String> {

    private static final Logger LOGGER
            = Logger.getLogger(BackupTask.class.getName());
    private static final ProcessExecutor PROCESS_EXECUTOR
            = WelcomeModelFactory.getProcessExecutor();

    private Partition exchangePartition;
    private Properties properties;

    private boolean backupConfigured;

    private BooleanProperty active = new SimpleBooleanProperty();
    private StringProperty sourcePath = new SimpleStringProperty();
    private BooleanProperty local = new SimpleBooleanProperty();
    private StringProperty destinationPath = new SimpleStringProperty();
    private BooleanProperty partition = new SimpleBooleanProperty();
    private StringProperty partitionPath = new SimpleStringProperty();
    private BooleanProperty screenshot = new SimpleBooleanProperty();
    private IntegerProperty frequency = new SimpleIntegerProperty();

    /**
     * Loads the backup configuration from the property file and the exchange
     * partition to preconfigure the destination path.
     *
     * @param properties Property File of the Welcome Application
     * @param backupDirectoryName the name for the backup folder
     */
    public BackupTask(Properties properties, String backupDirectoryName) {
        this.properties = properties;

        backupConfigured = "true".equals(properties.getProperty(
                WelcomeConstants.BACKUP));
        active.set(backupConfigured);
        sourcePath.set(properties.getProperty(
                WelcomeConstants.BACKUP_SOURCE, "/home/user/"));
        local.set("true".equals(properties.getProperty(
                WelcomeConstants.BACKUP_DIRECTORY_ENABLED, "true")));
        partition.set("true".equals(properties.getProperty(
                WelcomeConstants.BACKUP_PARTITION_ENABLED)));
        partitionPath.set(properties.getProperty(
                WelcomeConstants.BACKUP_PARTITION));
        screenshot.set("true".equals(properties.getProperty(
                WelcomeConstants.BACKUP_SCREENSHOT)));
        frequency.set(Integer.parseInt(properties.getProperty(
                WelcomeConstants.BACKUP_FREQUENCY, "5")));

        StorageDevice systemStorageDevice
                = WelcomeModelFactory.getSystemStorageDevice();
        if (systemStorageDevice != null) {
            exchangePartition = systemStorageDevice.getExchangePartition();

            if (exchangePartition != null) {
                try {
                    String exchangeMountPath = exchangePartition.getMountPath();
                    LOGGER.log(Level.INFO,
                            "exchangeMountPath: {0}", exchangeMountPath);
                    destinationPath.set(properties.getProperty(
                            WelcomeConstants.BACKUP_DIRECTORY,
                            exchangeMountPath + '/' + backupDirectoryName));
                } catch (DBusException | IOException ex) {
                    LOGGER.log(Level.SEVERE, "", ex);
                }
            }
        }
    }

    private boolean checkBackupDirectory() throws ProcessingException {
        if ((!active.get()) || (!local.get())) {
            // As long as the directory option is not selected we just don't
            // care what is configured there...
            return true;
        }

        String destination = destinationPath.get();

        if ((destination == null) || destination.isEmpty()) {
            throw new ProcessingException("Error_Title_Backup_Configuration",
                    "BackupTask.Error_No_Backup_Directory");
        }

        File dirFile = new File(destination);
        if (dirFile.exists()) {
            if (!dirFile.isDirectory()) {
                throw new ProcessingException(
                        "Error_Title_Backup_Configuration",
                        "BackupTask.Error_Backup_Directory_No_Directory",
                        destination);
            }

            String[] files = dirFile.list();
            if ((files != null) && (files.length != 0)) {
                int returnValue = PROCESS_EXECUTOR.executeProcess(
                        "rdiff-backup", "-l", dirFile.getAbsolutePath());
                if (returnValue != 0) {
                    throw new ProcessingException(
                            "Error_Title_Backup_Configuration",
                            "BackupTask.Error_Backup_Directory_Invalid",
                            destination);
                }
            }
        }

        // determine device where the directory is located
        // (df takes care for symlinks etc.)
        PROCESS_EXECUTOR.executeProcess(true, true, "df", destination);
        List<String> stdOut = PROCESS_EXECUTOR.getStdOutList();
        String device = null;
        for (String line : stdOut) {
            if (line.startsWith("/dev/")) {
                String[] tokens = line.split(" ");
                device = tokens[0];
            }
        }
        if (device == null) {
            LOGGER.log(Level.WARNING,
                    "could not determine device of directory {0}", destination);
            return true;
        }

        // check, if device is exFAT
        try {
            Partition destinationPartition
                    = Partition.getPartitionFromDeviceAndNumber(
                            device.substring(5));
            String idType = destinationPartition.getIdType();
            if (idType.equals("exfat")) {
                // rdiff-backup does not work (yet) on exfat partitions!
                throw new ProcessingException(
                        "Error_Title_Backup_Configuration",
                        "BackupTask.Error_Backup_on_exFAT", destination);
            }
        } catch (DBusException ex) {
            LOGGER.log(Level.WARNING, "", ex);
        }

        return true;
    }

    /**
     * Configures the backup by calling
     * {@link #updateJBackpackProperties(java.io.File, boolean)} for the
     * userPrefs directory of the user and the root.
     *
     * @param backupSource
     * @param backupDestination
     */
    private void updateJBackpackProperties(
            String backupSource, String backupDestination) {

        // update JBackpack preferences of the default user
        File prefsDirectory = new File(
                WelcomeConstants.USER_JBACKPACK_PREFERENCES);
        updateJBackpackProperties(prefsDirectory, backupSource,
                backupDestination, true);

        // update JBackpack preferences of the root user
        prefsDirectory = new File(WelcomeConstants.ROOT_JBACKPACK_PREFERENCES);
        updateJBackpackProperties(prefsDirectory, backupSource,
                backupDestination, false);
    }

    /**
     * Modifies the jbackpack xml file by changing the following values:
     * <ul>
     * <li>destination</li>
     * <li>local_destination_directory</li>
     * <li>source</li>
     * </ul>
     * If the xml couldn't be found it will be created by this function.
     *
     * @param prefsDirectory
     * @param backupSource
     * @param backupDestination
     * @param chown
     */
    private void updateJBackpackProperties(File prefsDirectory,
            String backupSource, String backupDestination, boolean chown) {
        File prefsFile = new File(prefsDirectory, "prefs.xml");
        String prefsFilePath = prefsFile.getPath();
        if (prefsFile.exists()) {
            try {
                Document xmlBootDocument = WelcomeUtil.parseXmlFile(prefsFile);
                xmlBootDocument.getDocumentElement().normalize();
                Node mapNode
                        = xmlBootDocument.getElementsByTagName("map").item(0);
                Element mapElement = (Element) mapNode;
                NodeList entries = mapElement.getElementsByTagName("entry");
                for (int i = 0, length = entries.getLength(); i < length; i++) {
                    Element entry = (Element) entries.item(i);
                    String key = entry.getAttribute("key");
                    switch (key) {
                        case "destination":
                            entry.setAttribute("value", "local");
                            break;
                        case "local_destination_directory":
                            entry.setAttribute("value", backupDestination);
                            break;
                        case "source":
                            entry.setAttribute("value", backupSource);
                            break;
                        default:
                            break;
                    }
                }

                // write changes back to config file
                File tmpFile = File.createTempFile("lernstickWelcome", "tmp");
                TransformerFactory transformerFactory
                        = TransformerFactory.newInstance();
                Transformer transformer = transformerFactory.newTransformer();
                transformer.setOutputProperty(OutputKeys.INDENT, "yes");
                transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
                        "http://java.sun.com/dtd/preferences.dtd");
                DOMSource source = new DOMSource(xmlBootDocument);
                StreamResult result = new StreamResult(tmpFile);
                transformer.transform(source, result);
                PROCESS_EXECUTOR.executeProcess(
                        "mv", tmpFile.getPath(), prefsFilePath);
                PROCESS_EXECUTOR.executeProcess(
                        "chown", "user.user", prefsFilePath);

            } catch (ParserConfigurationException | SAXException
                    | IOException | DOMException | TransformerException ex) {
                LOGGER.log(Level.WARNING, "can not update xmlboot config", ex);
            }

        } else {
            // JBackpack config doesn't exist and has to be created.
            if (!prefsDirectory.exists() && !prefsDirectory.mkdirs()) {
                LOGGER.log(Level.WARNING,
                        "could not create directory {0}", prefsDirectory);
                return;
            }
            // create minimal JBackpack preferences
            String preferences
                    = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n"
                    + "<!DOCTYPE map SYSTEM \"http://java.sun.com/dtd/preferences.dtd\">\n"
                    + "<map MAP_XML_VERSION=\"1.0\">\n"
                    + "  <entry key=\"destination\" value=\"local\"/>\n"
                    + "  <entry key=\"local_destination_directory\" value=\"" + destinationPath.get() + "\"/>\n"
                    + "  <entry key=\"source\" value=\"" + sourcePath.get() + "\"/>\n"
                    + "</map>\n";

            try (OutputStreamWriter osw = new OutputStreamWriter(
                    new FileOutputStream(prefsFile), Charset.defaultCharset())) {
                osw.write(preferences);
                osw.flush();
            } catch (IOException ex) {
                LOGGER.log(Level.WARNING, "", ex);
            }
            if (chown) {
                PROCESS_EXECUTOR.executeProcess(
                        "chown", "-R", "user.user", "/home/user/.java/");
            }
        }
    }

    public boolean isBackupConfigured() {
        return backupConfigured;
    }

    public boolean hasExchangePartition() {
        return exchangePartition != null;
    }

    public BooleanProperty activeProperty() {
        return active;
    }

    public StringProperty sourcePathProperty() {
        return sourcePath;
    }

    public BooleanProperty localProperty() {
        return local;
    }

    public StringProperty destinationPathProperty() {
        return destinationPath;
    }

    public BooleanProperty partitionProperty() {
        return partition;
    }

    public StringProperty partitionPathProperty() {
        return partitionPath;
    }

    public BooleanProperty screenshotProperty() {
        return screenshot;
    }

    public IntegerProperty frequencyProperty() {
        return frequency;
    }

    @Override
    public Task<String> newTask() {
        return new InternalTask();
    }

    /**
     * Task for {@link #newTask() }
     *
     * @see Processable
     */
    private class InternalTask extends Task<String> {

        @Override
        protected String call() throws Exception {
            updateProgress(0, 2);
            updateTitle("BackupTask.title");
            updateMessage("BackupTask.setupMessage");
            // XXX Move validation to controller/util to check it in GUI.
            if (checkBackupDirectory()) {
                if (!local.get()
                        || destinationPath.get() == null
                        || destinationPath.get().isEmpty()) {
                    if (partition.get()
                            && !partitionPath.get().isEmpty()) {
                        updateJBackpackProperties(sourcePath.get(), "/mnt/backup/"
                                + partitionPath.get() + "/lernstick_backup");
                    }
                } else {
                    updateJBackpackProperties(sourcePath.get(), destinationPath.get());
                }
                backupConfigured = true;
            }
            updateProgress(1, 2);
            updateMessage("BackupTask.saveConfig");
            properties.setProperty(WelcomeConstants.BACKUP,
                    active.get() ? "true" : "false");
            properties.setProperty(WelcomeConstants.BACKUP_SCREENSHOT,
                    screenshot.get() ? "true" : "false");
            properties.setProperty(WelcomeConstants.BACKUP_DIRECTORY_ENABLED,
                    local.get() ? "true" : "false");
            properties.setProperty(WelcomeConstants.BACKUP_PARTITION_ENABLED,
                    partition.get() ? "true" : "false");
            properties.setProperty(WelcomeConstants.BACKUP_SOURCE,
                    sourcePath.get() != null ? sourcePath.get() : "");
            properties.setProperty(WelcomeConstants.BACKUP_DIRECTORY,
                    destinationPath.get() != null ? destinationPath.get() : "");
            properties.setProperty(WelcomeConstants.BACKUP_PARTITION,
                    partitionPath.get() != null ? partitionPath.get() : "");
            properties.setProperty(WelcomeConstants.BACKUP_FREQUENCY,
                    Integer.toString(frequency.get()));
            updateProgress(2, 2);
            return null;
        }
    }

}
